@@ -35,10 +35,16 @@
     Promise,
     PromisePrototypeThen,
     PromisePrototypeCatch,
+    String,
     StringPrototypeToLowerCase,
     TypedArrayPrototypeSubarray,
     TypeError,
     Uint8Array,
+    WeakMap,
+    WeakMapPrototypeDelete,
+    WeakMapPrototypeGet,
+    WeakMapPrototypeHas,
+    WeakMapPrototypeSet,
   } = window.__bootstrap.primordials;
 
   const REQUEST_BODY_HEADER_NAMES = [
@@ -48,6 +54,8 @@
     "content-type",
   ];
 
+  const requestBodyReaders = new WeakMap();
+
   /**
    * @param {{ method: string, url: string, headers: [string, string][], clientRid: number | null, hasBody: boolean }} args
    * @param {Uint8Array | null} body
@@ -85,11 +93,7 @@
 
   // A finalization registry to clean up underlying fetch resources that are GC'ed.
   const RESOURCE_REGISTRY = new FinalizationRegistry((rid) => {
-    try {
-      core.close(rid);
-    } catch {
-      // might have already been closed
-    }
+    core.tryClose(rid);
   });
 
   /**
@@ -105,11 +109,7 @@
           new DOMException("Ongoing fetch was aborted.", "AbortError"),
         );
       }
-      try {
-        core.close(responseBodyRid);
-      } catch (_) {
-        // might have already been closed
-      }
+      core.tryClose(responseBodyRid);
     }
     // TODO(lucacasonato): clean up registration
     terminator[abortSignal.add](onAbort);
@@ -131,11 +131,7 @@
             RESOURCE_REGISTRY.unregister(readable);
             // We have reached the end of the body, so we close the stream.
             controller.close();
-            try {
-              core.close(responseBodyRid);
-            } catch (_) {
-              // might have already been closed
-            }
+            core.tryClose(responseBodyRid);
           }
         } catch (err) {
           RESOURCE_REGISTRY.unregister(readable);
@@ -148,11 +144,7 @@
             // error.
             controller.error(err);
           }
-          try {
-            core.close(responseBodyRid);
-          } catch (_) {
-            // might have already been closed
-          }
+          core.tryClose(responseBodyRid);
         }
       },
       cancel() {
@@ -172,6 +164,33 @@
    * @returns {Promise<InnerResponse>}
    */
   async function mainFetch(req, recursive, terminator) {
+    if (req.blobUrlEntry !== null) {
+      if (req.method !== "GET") {
+        throw new TypeError("Blob URL fetch only supports GET method.");
+      }
+
+      const body = new InnerBody(req.blobUrlEntry.stream());
+      terminator[abortSignal.add](() =>
+        body.error(new DOMException("Ongoing fetch was aborted.", "AbortError"))
+      );
+
+      return {
+        headerList: [
+          ["content-length", String(req.blobUrlEntry.size)],
+          ["content-type", req.blobUrlEntry.type],
+        ],
+        status: 200,
+        statusMessage: "OK",
+        body,
+        type: "basic",
+        url() {
+          if (this.urlList.length == 0) return null;
+          return this.urlList[this.urlList.length - 1];
+        },
+        urlList: recursive ? [] : [...req.urlList],
+      };
+    }
+
     /** @type {ReadableStream<Uint8Array> | Uint8Array | null} */
     let reqBody = null;
 
@@ -181,6 +200,7 @@
           reqBody = req.body.stream;
         } else {
           const reader = req.body.stream.getReader();
+          WeakMapPrototypeSet(requestBodyReaders, req, reader);
           const r1 = await reader.read();
           if (r1.done) {
             reqBody = new Uint8Array(0);
@@ -189,6 +209,7 @@
             const r2 = await reader.read();
             if (!r2.done) throw new TypeError("Unreachable");
           }
+          WeakMapPrototypeDelete(requestBodyReaders, req);
         }
       } else {
         req.body.streamOrStatic.consumed = true;
@@ -206,15 +227,11 @@
     }, reqBody instanceof Uint8Array ? reqBody : null);
 
     function onAbort() {
-      try {
-        core.close(cancelHandleRid);
-      } catch (_) {
-        // might have already been closed
+      if (cancelHandleRid !== null) {
+        core.tryClose(cancelHandleRid);
       }
-      try {
-        core.close(requestBodyRid);
-      } catch (_) {
-        // might have already been closed
+      if (requestBodyRid !== null) {
+        core.tryClose(requestBodyRid);
       }
     }
     terminator[abortSignal.add](onAbort);
@@ -224,6 +241,7 @@
         throw new TypeError("Unreachable");
       }
       const reader = reqBody.getReader();
+      WeakMapPrototypeSet(requestBodyReaders, req, reader);
       (async () => {
         while (true) {
           const { value, done } = await PromisePrototypeCatch(
@@ -252,11 +270,8 @@
             break;
           }
         }
-        try {
-          core.close(requestBodyRid);
-        } catch (_) {
-          // might have already been closed
-        }
+        WeakMapPrototypeDelete(requestBodyReaders, req);
+        core.tryClose(requestBodyRid);
       })();
     }
 
@@ -267,10 +282,8 @@
         throw err;
       });
     } finally {
-      try {
-        core.close(cancelHandleRid);
-      } catch (_) {
-        // might have already been closed
+      if (cancelHandleRid !== null) {
+        core.tryClose(cancelHandleRid);
       }
     }
     if (terminator.aborted) return abortedNetworkError();
@@ -471,7 +484,13 @@
 
   function abortFetch(request, responseObject) {
     const error = new DOMException("Ongoing fetch was aborted.", "AbortError");
-    if (request.body !== null) request.body.cancel(error);
+    if (request.body !== null) {
+      if (WeakMapPrototypeHas(requestBodyReaders, request)) {
+        WeakMapPrototypeGet(requestBodyReaders, request).cancel(error);
+      } else {
+        request.body.cancel(error);
+      }
+    }
     if (responseObject !== null) {
       const response = toInnerResponse(responseObject);
       if (response.body !== null) response.body.error(error);
@@ -486,8 +505,8 @@
    *
    * @param {any} source The source parameter that the WebAssembly
    * streaming API was called with.
-   * @param {number} rid An rid that can be used with
-   * `Deno.core.wasmStreamingFeed`.
+   * @param {number} rid An rid that represents the wasm streaming
+   * resource.
    */
   function handleWasmStreaming(source, rid) {
     // This implements part of
@@ -524,15 +543,15 @@
           while (true) {
             const { value: chunk, done } = await reader.read();
             if (done) break;
-            core.wasmStreamingFeed(rid, "bytes", chunk);
+            core.opSync("op_wasm_streaming_feed", rid, chunk);
           }
         }
 
         // 2.7.
-        core.wasmStreamingFeed(rid, "finish");
+        core.close(rid);
       } catch (err) {
         // 2.8 and 3
-        core.wasmStreamingFeed(rid, "abort", err);
+        core.opSync("op_wasm_streaming_abort", rid, err);
       }
     })();
   }
